"""
Triplet generators.

Functions for creating generators that will yield batches of triplets.

Triplets will be created using neighbour matrices, which can either be precomputed or
dynamically generated.
"""

from abc import ABC, abstractmethod
import numpy as np
from tensorflow.keras.utils import Sequence
from scipy.sparse import issparse


def generator_from_neighbour_matrix(X, Y, neighbour_matrix, batch_size):
    if Y is None:
        return UnsupervisedTripletGenerator(X, neighbour_matrix, batch_size=batch_size)

    return SupervisedTripletGenerator(X, Y, neighbour_matrix, batch_size=batch_size)


class TripletGenerator(Sequence, ABC):
    def __init__(self, X, neighbour_matrix, batch_size=32):
        if batch_size > X.shape[0]:
            raise Exception('''batch_size value larger than num_rows in dataset
                            (batch_size={}, rows={}). Lower batch_size to a
                            smaller value.'''.format(batch_size, X.shape[0]))
        self.X = X
        self.neighbour_matrix = neighbour_matrix
        self.batch_size = batch_size

    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_indices = range(idx * self.batch_size,
                              min((idx + 1) * self.batch_size, self.X.shape[0]))

        label_batch = self.get_labels(batch_indices)
        triplet_batch = [self.get_triplet(row_index)
                         for row_index in batch_indices]

        if issparse(self.X):
            triplet_batch = [[e.toarray()[0] for e in t] for t in triplet_batch]

        triplet_batch = np.array(triplet_batch)

        return self.output_triplets(triplet_batch, label_batch)

    def get_triplet(self, idx):
        triplet = []
        neighbour_list = self.get_neighbours(idx)
        neighbour_list = np.array(neighbour_list, dtype=np.uint32)

        # Take a random neighbour as positive
        neighbour_ind = np.random.choice(neighbour_list)

        # Take a random non-neighbour as negative
        # Pick a random index until one fits constraint, usually faster.
        negative_ind = np.random.randint(0, self.X.shape[0])
        while negative_ind in neighbour_list:
            negative_ind = np.random.randint(0, self.X.shape[0])

        triplet += [self.X[idx], self.X[neighbour_ind], self.X[negative_ind]]
        return triplet

    def get_neighbours(self, idx):
        return self.neighbour_matrix[idx]

    @abstractmethod
    def get_labels(self, batch_indices):
        raise NotImplementedError("Override this method with a concrete implementation")

    @abstractmethod
    def output_triplets(self, triplet_batch, label_batch):
        raise NotImplementedError("Override this method with a concrete implementation")


class UnsupervisedTripletGenerator(TripletGenerator):
    def __init__(self, X, neighbour_matrix, batch_size=32):
        super().__init__(X, neighbour_matrix, batch_size)
        self.placeholder_labels = np.empty(batch_size, dtype=np.uint8)

    def get_labels(self, batch_indices):
        return self.placeholder_labels[:len(batch_indices)]

    def output_triplets(self, triplet_batch, label_batch):
        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), label_batch


class SupervisedTripletGenerator(TripletGenerator):
    def __init__(self, X, Y, neighbour_matrix, batch_size=32):
        super().__init__(X, neighbour_matrix, batch_size)
        self.Y = Y

    def get_labels(self, batch_indices):
        return self.Y[batch_indices]

    def output_triplets(self, triplet_batch, label_batch):
        label_batch = np.array(label_batch)
        return tuple([triplet_batch[:, 0], triplet_batch[:, 1], triplet_batch[:, 2]]), tuple([label_batch, label_batch])



""" KNN retrieval using an Annoy index. """

from contextlib import suppress
import functools
import shutil
import time
from multiprocessing import Array, Process, Value, cpu_count, Queue
from collections.abc import Sequence
from operator import attrgetter
from scipy.sparse import issparse
from annoy import AnnoyIndex
from tqdm import tqdm
import numpy as np


class AnnoyKnnMatrix(Sequence):
    r"""A matrix |A_ij| where i is the row index of the data point and j
    refers to the index of the neigbouring point. Neighbouring points
    are KNN retrieved using an Annoy Index.

    .. |A_ij| replace:: A\ :subscript:`ij`

    :param AnnoyIndex index: AnnoyIndex instance to use when retrieving KNN
    :param tuple shape: Shape of the KNN matrix (height, width)
    :param string index_path: Location of the AnnoyIndex file on disk
    :param int k: The number of neighbours to retrieve for each point
    :param int search_k: Controls the number of nodes searched - higher is more
        expensive but more accurate. Default of -1 defaults to n_trees * k
    :param boolean precompute: Whether to precompute the KNN index and store the matrix in memory.
        Much faster when training, but consumes more memory.
    :param boolean include_distances: Whether to return the distances along with the indexes of
        the neighbouring points
    :param boolean verbose: Controls verbosity of output to console when building index. If
        False, nothing will be printed to the terminal."""

    def __init__(self, index, shape, index_path='annoy.index', k=150, search_k=-1,
                 precompute=False, include_distances=False, verbose=False):
        """Constructs an AnnoyKnnMatrix instance from an AnnoyIndex object with given parameters"""
        self.index = index
        self.shape = shape
        self.index_path = index_path
        self.k = k
        self.search_k = search_k
        self.include_distances = include_distances
        self.verbose = verbose
        self.precomputed_neighbours = None
        if precompute:
            self.precomputed_neighbours = self.get_neighbour_indices()

    def unload(self):
        """Unloads the index from disk, allowing other processes to read/write to the index file.
        After calling this, the index will no longer be usable from this instance."""
        self.index.unload()

    @classmethod
    def build(cls, X, path, k=150, metric='angular', search_k=-1, include_distances=False,
              ntrees=50, build_index_on_disk=True, precompute=False, verbose=1):
        """Builds a new Annoy Index on input data *X*, then constructs and returns an
        AnnoyKnnMatrix object using the newly-built index."""

        _validate_knn_shape(X.shape, k)
        index = build_annoy_index(X, path, metric, ntrees, build_index_on_disk, verbose)
        return cls(index, X.shape, path, k, search_k, precompute, include_distances, verbose)

    @classmethod
    def load(cls, index_path, shape, k=150, search_k=-1, include_distances=False,
             precompute=False, verbose=1):
        """Constructs and returns an AnnoyKnnMatrix object from an existing Annoy Index on disk."""

        _validate_knn_shape(shape, k)
        index = AnnoyIndex(shape[1], metric='angular')
        index.load(index_path)
        return cls(index, shape, index_path, k, search_k, precompute, include_distances, verbose)

    def __len__(self):
        """Number of rows in neighbour matrix"""
        return self.shape[0]

    def __getitem__(self, idx):
        """Returns neighbours list for the specified index.
        Supports both integer and slice indices."""
        if isinstance(idx, slice):
            start, stop, step = idx.indices(len(self))
            return [self[i] for i in range(start, stop, step)]
        if idx < 0:
            idx += len(self)
        if self.precomputed_neighbours is not None:
            return self.precomputed_neighbours[idx]
        return self.index.get_nns_by_item(
            idx, self.k, search_k=self.search_k, include_distances=self.include_distances)

    def get_neighbour_indices(self):
        """Retrieves neighbours for every row in parallel"""
        if self.precomputed_neighbours is not None:
            return self.precomputed_neighbours
        return extract_knn(
            self.shape, k=self.k,
            index_builder=self.load,
            verbose=self.verbose,
            index_path=self.index_path,
            search_k=self.search_k,
            include_distances=self.include_distances)

    def save(self, path):
        """Saves internal Annoy index to disk at given path."""
        self.index.save(path)

def cleanup_knn_index(index, path):
    """Cleans up disk resources used by a KNN index.
    First will attempt to call the `unload` method on the index,
    then will recursively remove the files at the path provided."""
    with suppress(Exception):
        index.unload()
    shutil.rmtree(path, ignore_errors=True)

def _validate_knn_shape(shape, k):
    if k >= shape[0]:
        raise ValueError('''k value greater than or equal to num_rows
                            (k={}, rows={}). Lower k to a smaller
                            value.'''.format(k, shape[0]))

def build_annoy_index(X, path, metric='angular', ntrees=50, build_index_on_disk=True, verbose=1):
    """ Build a standalone annoy index.

    :param array X: numpy array with shape (n_samples, n_features)
    :param str path: The filepath of a trained annoy index file
        saved on disk.
    :param int ntrees: The number of random projections trees built by Annoy to
        approximate KNN. The more trees the higher the memory usage, but the
        better the accuracy of results.
    :param bool build_index_on_disk: Whether to build the annoy index directly
        on disk. Building on disk should allow for bigger datasets to be indexed,
        but may cause issues.
    :param str metric: Which distance metric Annoy should use when building KNN index.
        Supports "angular", "euclidean", "manhattan", "hamming", or "dot".
    :param int verbose: Controls the volume of logging output the model
        produces when training. When set to 0, silences outputs, when above 0
        will print outputs.

    """

    if verbose:
        print("Building KNN index")

    if len(X.shape) > 2:
        if not "reshape" in dir(X):
            raise ValueError("Attempting to build AnnoyIndex on multi-dimensional data"
                             " without providing a reshape method. AnnoyIndexes require"
                             " 2D data - rows and columns.")
        if verbose:
            print('Flattening multidimensional input before building KNN index using Annoy')
        X = X.reshape((X.shape[0], -1))

    index = AnnoyIndex(X.shape[1], metric=metric)
    if build_index_on_disk:
        index.on_disk_build(path)

    if issparse(X):
        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
            vector = X[i].toarray()[0]
            index.add_item(i, vector)
    else:
        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
            vector = X[i]
            index.add_item(i, vector)

    try:
        index.build(ntrees)
    except Exception as e:
        msg = ("Error building Annoy Index. Passing on_disk_build=False"
               " may solve the issue, especially on Windows.")
        raise Exception(msg) from e

    if not build_index_on_disk:
        index.save(path)
    return index

def extract_knn(data_shape, k, index_builder=AnnoyKnnMatrix.load, verbose=1, **kwargs):
    """Starts multiple processes to retrieve nearest neighbours from a built index in parallel.

    :param int k: Number of neighbours to retrieve. Passed as a kwarg to index builder function.
    :param tuple data_shape: The shape of the data that the index was built on.
    :param Callable index_builder: Called within each worker process to load the index.
    :param int verbose: Controls verbosity of output.
    :param kwargs: keyword arguments to pass to the index_builder function."""

    if verbose:
        print("Extracting KNN neighbours")

    if len(data_shape) > 2:
        if verbose:
            print('Flattening data before retrieving KNN from index')
        data_shape = (data_shape[0], functools.reduce(lambda x, y: x*y, data_shape[1:]))

    array_ctype = _get_uint_ctype(data_shape[0])
    neighbours_array = Array(array_ctype, data_shape[0] * k, lock=False)

    chunk_size = max(1, data_shape[0] // cpu_count())
    process_pool = []
    progress_counters = []
    error_queue = Queue()

    # Split up the indices and assign processes for each chunk
    for i in range(0, data_shape[0], chunk_size):
        counter = Value(array_ctype, 0, lock=False)
        process_pool.append(KnnWorker(neighbours_array, (i, min(i+chunk_size, data_shape[0])),
                                      index_builder, k=k, shape=data_shape, error_queue=error_queue,
                                      progress_counter=counter, **kwargs))
        progress_counters.append(counter)
    try:
        for process in process_pool:
            process.start()

        # Poll for progress updates
        with tqdm(total=data_shape[0], disable=verbose < 1) as pbar:
            while pbar.n < data_shape[0]:
                # Raise worker errors
                if not error_queue.empty():
                    raise error_queue.get()

                num_processed = sum([num.value for num in progress_counters])
                pbar.update(num_processed - pbar.n)
                time.sleep(0.1)

        # Join processes to avoid zombie processes on UNIX
        for process in process_pool:
            process.join()

        neighbour_matrix = np.ndarray((data_shape[0], k), buffer=neighbours_array,
                                      dtype=array_ctype)
        return neighbour_matrix

    except:
        print('Halting KNN retrieval and cleaning up')
        for process in process_pool:
            process.terminate()
            process.join()
        raise


class KnnWorker(Process):
    """When run, this worker process loads an index from disk using the provided
    'knn_init' function, passing all additional 'kwargs' to this initializer function.
    Then the neighbours of the data-points between the range specified by `data_indices`
    are retrieved and stored in the 'neighbours_array'.

    :param: multiprocessing.Array neighbours_array. 1D array that worker will insert results into.
    :param: tuple data_indices. Specifies range (start, end) to retrieve neighbours for.
    :param: Callable knn_init. Function to load index from disk.
    :param: int k. Number of neighbours to retrieve. Will be passed as kwargs to knn_init function.
    :param: Optional[multiprocessing.Value] progress_counter. Incremented for every row processed.
    :param: Optional[multiprocessing.Queue] error_queue. Exceptions encountered are sent to this
        queue if provided.
    :param: **kwargs. Args to pass to knn_init function.
    """

    def __init__(self, neighbours_array, data_indices, knn_init, k,
                 progress_counter=None, error_queue=None, **kwargs):
        super().__init__()
        self.neighbours_array = neighbours_array
        self.data_indices = data_indices
        self.k = k
        self.knn_init = knn_init
        self.progress_counter = progress_counter
        self.error_queue = error_queue
        self.kwargs = kwargs
        self.kwargs['k'] = k

    def run(self):
        """Load index, retrieve neighbouring points, and insert results into 1D array.
        Can update progress_counter and send Exceptions to error queue if provided."""
        try:
            knn_index = self.knn_init(**self.kwargs)

            for i in range(self.data_indices[0], self.data_indices[1]):
                neighbour_indexes = knn_index[i]
                row_offset = i * self.k
                for j in range(0, self.k):
                    self.neighbours_array[row_offset + j] = neighbour_indexes[j]
                if self.progress_counter is not None:
                    self.progress_counter.value += 1
        except Exception as e:
            if self.error_queue:
                self.error_queue.put(e)
            raise

def _get_uint_ctype(integer):
    """Gets smallest possible uint representation of provided integer.
    Raises ValueError if invalid value provided (negative or above uint64)."""
    for dtype in np.typecodes["UnsignedInteger"]:
        min_val, max_val = attrgetter('min', 'max')(np.iinfo(np.dtype(dtype)))
        if min_val <= integer <= max_val:
            return dtype
    raise ValueError("Cannot parse {} as a uint64".format(integer))



"""Generators for non-triplet data"""

import tensorflow as tf
import numpy as np

from scipy.sparse import issparse


class KerasSequence(tf.keras.utils.Sequence):
    """Wraps inputs into a Keras Sequence to allow Keras models to predict on
    arbitrary inputs which may be out of memory."""
    def __init__(self, X, batch_size=32):
        self.X = X
        self.batch_size = batch_size
        self.placeholder_labels = np.empty(batch_size, dtype=np.uint8)
    def __len__(self):
        return int(np.ceil(self.X.shape[0] / float(self.batch_size)))
    def __getitem__(self, index):
        batch_indices = range(index * self.batch_size, min((index + 1) * self.batch_size, self.X.shape[0]))

        if issparse(self.X):
            batch = np.array([self.X[i].toarray() for i in batch_indices])
            batch = np.squeeze(batch)
        else:
            batch = np.array([self.X[i] for i in batch_indices])
        placeholder_labels = self.placeholder_labels[:len(batch_indices)]
        return batch, placeholder_labels



"""A collection of callbacks that can be passed to ivis to be called during training.
These provide utilities such as saving checkpoints during training (allowing for
resuming if interrupted), as well as periodic logging of plots and model embeddings.
With this information, you may decide to terminate a training session early
due to a lack of improvements to the visualizations, for example.

To use a callback during training, simply pass a list of callback objects
to the Ivis object when creating it using the callbacks keyword argument.
The ivis.nn.callbacks module contains a set of callbacks provided for use with ivis models,
but any `tf.keras.callbacks.Callbacks` object can be passed and will be used during training:
for example, `tf.keras.callbacks.TensorBoard`. This means it's also possible to write your own
callbacks for ivis to use.
"""

import io
import os
import numpy as np
import tensorflow as tf
from tensorflow.keras.callbacks import Callback
from sklearn.preprocessing import MinMaxScaler

# Matplotlib and seaborn are optional dependencies
try:
    from matplotlib import pyplot as plt
except ImportError:
    plt = None
try:
    import seaborn as sns
except ImportError:
    sns = None


class ModelCheckpoint(Callback):
    """Periodically saves the model during training. By default, it saves the
    model every epoch; increasing the ``epoch_interval`` will make
    checkpointing less frequent.

    If the given filename contains the ``{}`` string, the epoch number will be
    subtituted in, resulting in multiple checkpoint folders with different
    names. If a filename such as 'ivis-checkpoint' is provided, only the
    latest checkpoint will be kept.

    :param str log_dir: Folder to save resulting embeddings.
    :param str filename: Filename to save each file as. `{}` in string
        will be substituted with the epoch number.

    Example usage:
    ::

        from ivis.nn.callbacks import ModelCheckpoint
        from ivis import Ivis

        # Save only the latest checkpoint to current directory every 10 epochs
        checkpoint_callback = ModelCheckpoint(log_dir='.',
                                            filename='latest-checkpoint.ivis',
                                            epoch_interval=10)

        model = Ivis(callbacks=[checkpoint_callback])
    """

    def __init__(self, log_dir='./model_checkpoints',
                 filename='model-checkpoint_{}.ivis', epoch_interval=1):
        super(ModelCheckpoint, self).__init__()
        self.log_dir = log_dir
        os.makedirs(log_dir, exist_ok=True)
        self.filename = filename
        self.epoch_interval = epoch_interval
        self.epochs_since_last_save = 0
        self.ivis_model = None

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.epoch_interval:
            self.epochs_since_last_save = 0
            filename = self.filename.format(epoch + 1)
            self.ivis_model.save_model(
                os.path.join(self.log_dir, filename), overwrite=True)

    def register_ivis_model(self, ivis_model):
        self.ivis_model = ivis_model


class EmbeddingsLogging(Callback):
    """Periodically saves embeddings of the data provided to ``data``
    using the latest state of the ``Ivis`` model.
    By default, saves embeddings every epoch; increasing the
    ``epoch_interval`` will save the embeddings less frequently.

    :param list[float] data: Data to embed with the latest Ivis object
    :param str log_dir: Folder to save resulting embeddings.
    :param str filename: Filename to save each file as. `{}` in string
        will be substituted with the epoch number.

    Example usage:
    ::

        from ivis.nn.callbacks import EmbeddingsLogging
        from ivis import Ivis
        from tensorflow.keras.datasets import mnsit

        (X_train, Y_train), (X_test, Y_test)  = mnist.load_data()

        # Save embeddings of test set every epoch
        embeddings_callback = EmbeddingsLogging(X_test,
                                                log_dir='test-embeddings',
                                                filename='{}_test_embeddings.npy',
                                                epoch_interval=1)

        model = Ivis(callbacks=[embeddings_callback])

        # Train on training set
        model.fit(X_train)
    """

    def __init__(self, data, log_dir='./embeddings_logs',
                 filename='{}_embeddings.npy', epoch_interval=1):
        super(EmbeddingsLogging, self).__init__()
        self.data = data
        self.log_dir = log_dir
        os.makedirs(log_dir, exist_ok=True)
        self.filename = filename
        self.epoch_interval = epoch_interval
        self.epochs_since_last_save = 0
        self.embeddings = None

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.epoch_interval:
            self.epochs_since_last_save = 0
            self.embeddings = self.model.layers[3].predict(self.data)
            filename = self.filename.format(epoch + 1)
            np.save(os.path.join(self.log_dir, filename), self.embeddings)


class EmbeddingsImage(Callback):
    """Periodically generates and plots 2D embeddings of the data
    provided to ``data`` using the latest state of the ``Ivis`` model.
    By default, saves plots of the embeddings every epoch; increasing the
    ``epoch_interval`` will save the plots less frequently.

    :param list[float] data: Data to embed and plot with the latest Ivis model
    :param list[int] labels: Labels with which to colour plotted
        embeddings. If `None` all points will have the same color.
    :param str log_dir: Folder to save resulting embeddings.
    :param str filename: Filename to save each file as. `{}` in string
        will be substituted with the epoch number.

    Example usage:
    ::

        from ivis.nn.callbacks import EmbeddingsImage
        from ivis import Ivis
        from tensorflow.keras.datasets import mnsit

        (X_train, Y_train), (X_test, Y_test)  = mnist.load_data()

        # Plot embeddings of test set every epoch colored by labels
        embeddings_callback = EmbeddingsImage(X_test, Y_test,
                                                log_dir='test-embeddings',
                                                filename='{}_test_embeddings.npy',
                                                epoch_interval=1)

        model = Ivis(callbacks=[embeddings_callback])

        # Train on training set
        model.fit(X_train)
    """

    def __init__(self, data, labels=None, log_dir='./logs',
                 filename='{}_embeddings.png', epoch_interval=1):
        super(EmbeddingsImage, self).__init__()
        _check_visualization_libraries()

        self.data = data
        if labels is not None:
            self.labels = labels
        else:
            self.labels = np.zeros(len(data))
        self.n_classes = len(np.unique(self.labels, axis=0))
        self.log_dir = log_dir
        os.makedirs(log_dir, exist_ok=True)
        self.filename = filename
        self.epoch_interval = epoch_interval
        self.epochs_since_last_save = 0
        self.embeddings = None

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.epoch_interval:
            self.epochs_since_last_save = 0
            self.embeddings = self.model.layers[3].predict(self.data)
            filename = self.filename.format(epoch + 1)
            self.plot_embeddings(filename)

    def plot_embeddings(self, filename):
        embeddings = MinMaxScaler((0, 1)).fit_transform(self.embeddings)

        fig = plt.figure()
        sns.scatterplot(x=embeddings[:, 0], y=embeddings[:, 1], s=1,
                        hue=self.labels,
                        palette=sns.color_palette("hls", self.n_classes),
                        linewidth=0)

        plt.savefig(os.path.join(self.log_dir, filename), dpi=300)
        plt.close(fig)


class TensorBoardEmbeddingsImage(Callback):
    """Periodically generates and plots 2D embeddings of the data
    provided to ``data`` using the latest state of the ``Ivis`` model.
    The plots are designed to be viewed in Tensorboard, which will provide
    an image that shows the history of embeddings plots through training.
    By default, saves plots of the embeddings every epoch; increasing the
    ``epoch_interval`` will save the plots less frequently.

    :param list[float] data: Data to embed and plot with the latest Ivis
    :param list[int] labels: Labels with which to colour plotted
        embeddings. If `None` all points will have the same color.
    :param str log_dir: Folder to save resulting embeddings.
    :param str filename: Filename to save each file as. `{}` in string
        will be substituted with the epoch number.

    Example usage:
    ::

        from ivis.nn.callbacks import TensorBoardEmbeddingsImage
        from ivis import Ivis
        from tensorflow.keras.datasets import mnsit

        (X_train, Y_train), (X_test, Y_test)  = mnist.load_data()

        # Plot embeddings of test set every epoch colored by labels
        embeddings_callback = TensorBoardEmbeddingsImage(X_test, Y_test,
                                                log_dir='test-embeddings',
                                                filename='{}_test_embeddings.npy',
                                                epoch_interval=1)

        model = Ivis(callbacks=[embeddings_callback])

        # Train on training set
        model.fit(X_train)
    """

    def __init__(self, data, labels=None,
                 log_dir='./logs', epoch_interval=1):
        super(TensorBoardEmbeddingsImage, self).__init__()
        _check_visualization_libraries()

        self.data = data
        self.log_dir = log_dir
        if labels is not None:
            self.labels = labels
        else:
            self.labels = np.zeros(len(data))
        self.n_classes = len(np.unique(self.labels, axis=0))
        self.epochs_since_last_save = 0
        self.epoch_interval = epoch_interval
        self.embeddings = None
        self.file_writer = tf.summary.create_file_writer(
            os.path.join(log_dir, 'embeddings')
        )

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.epoch_interval:
            self.embeddings = self.model.layers[3].predict(self.data)
            image = self.plot_embeddings(self.embeddings)
            with self.file_writer.as_default():
                tf.summary.image("Embeddings", image, step=epoch)

    def plot_embeddings(self, embeddings):
        embeddings = MinMaxScaler((0, 1)).fit_transform(self.embeddings)

        fig = plt.figure()
        buf = io.BytesIO()
        sns.scatterplot(x=embeddings[:, 0], y=embeddings[:, 1], s=1,
                        hue=self.labels,
                        palette=sns.color_palette("hls", self.n_classes),
                        linewidth=0)

        plt.savefig(buf, format='png', dpi=300)
        plt.close(fig)
        buf.seek(0)

        image = tf.image.decode_png(buf.getvalue(), channels=4)
        image = tf.expand_dims(image, 0)
        return image


def _check_visualization_libraries():
    if plt is None:
        raise ImportError(
            'Failed to import `matplotlib`.'
            'To use visualization callbacks install `matplotlib`.'
            'For example: pip install matplotlib'
        )
    if sns is None:
        raise ImportError(
            'Failed to import `seaborn`.'
            'To use visualization callbacks install `seaborn`.'
            'For example: pip install seaborn'
        )



"""Common interface for all custom datasets compatible with ivis"""

from abc import abstractmethod
from collections.abc import Sequence

class IndexableDataset(Sequence):
    """A sequence that also defines a shape attribute. This indexable data structure
    can be provided as input to ivis."""
    @abstractmethod
    def shape(self):
        """Returns the shape of the dataset. First dimension corresponds to rows,
        the other dimensions correspond to features."""
        raise NotImplementedError



"""Custom datasets that load images from disk."""

import numpy as np
import tensorflow as tf

from .sequence import IndexableDataset


class ImageDataset(IndexableDataset):
    """When indexed, loads images from disk, resizes to consistent size, then returns image.
    Since the returned images will consist of 3 dimensions, the model ivis uses must be capable
    of dealing with this dimensionality of data (for example, a Convolutional Neural Network).
    Such a model can be constructed externally and then passed to ivis as the argument for 'model'.

    :param list filepath_list: All image filepaths in dataset.
    :param tuple img_shape: A tuple (height, width) containing desired dimensions to resize the
        images to.
    :param color_mode str: Either "rgb", "rgba" or "grayscale". Determines how many channels present
        in images that are read in - 3, 4, or 1 respectively.
    :param resize_method str: Interpolation method to use when resizing image. Must be one of:
        "area", "bicubic", "bilinear", "gaussian", "lanczos3", "lanczos5", "mitchellcubic", "nearest".
    :param preserve_aspect_ratio boolean: Whether to preserve the aspect ratio when resizing images.
        If True, will maintain aspect ratio by padding the image.
    :param dtype tf.dtypes.DType: The dtype to read the image into. One of tf.uint8 or tf.uint16.
    :param preprocessing_function Callable: A function to apply to every image. Will be called
        at the end of the pipeline, after image reading and resizing.
        If None (default), no function will be applied."""

    shape = None
    def __init__(self, filepath_list, img_shape, color_mode="rgb",
                 resize_method="bilinear", preserve_aspect_ratio=False,
                 dtype=tf.uint8, preprocessing_function=None):
        self.filepath_list = np.array(filepath_list)
        self.img_shape = img_shape
        if color_mode == "rgb":
            self.channels = 3
        elif color_mode == "rgba":
            self.channels = 4
        elif color_mode == "grayscale":
            self.channels = 1
        else:
            raise ValueError("color_mode arg '{}' not recognized."
                             " Must be one of 'rgb', 'rgba', 'grayscale.".format(color_mode))
        self.resize_method = resize_method
        self.preserve_aspect_ratio = preserve_aspect_ratio
        self.dtype = dtype
        self.preprocessing_function = preprocessing_function
        self.shape = (len(filepath_list), *img_shape, self.channels)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            start, stop, step = idx.indices(len(self))
            return [self[i] for i in range(start, stop, step)]
        if idx < 0:
            idx += len(self)
        img = self.read_image(self.filepath_list[idx])
        img = self.resize_image(img)
        if self.preprocessing_function:
            img = self.preprocessing_function(img)
        return img

    def __len__(self):
        return len(self.filepath_list)

    def read_image(self, filepath):
        """Reads an image from disk into a numpy array"""
        img_bytes = tf.io.read_file(filepath)
        return tf.image.decode_png(img_bytes, channels=self.channels, dtype=self.dtype)

    def resize_image(self, img):
        """Resizes an numpy array image to desired dimensions"""
        if not self.preserve_aspect_ratio:
            img = tf.image.resize(img, self.img_shape, method=self.resize_method)
        else:
            img = tf.image.resize_with_pad(img, *self.img_shape, method=self.resize_method)
        return tf.cast(img, self.dtype).numpy()


class FlattenedImageDataset(ImageDataset):
    """Returns flattened versions of images read in from disk. This dataset can be
    used with the default neighbour retrieval method used by ivis (Annoy KNN index)
    since it is 2D."""

    def __init__(self, filepath_list, img_shape, color_mode="rgb",
                 resize_method="bilinear", preserve_aspect_ratio=False,
                 dtype=tf.uint8, preprocessing_function=None):
        super().__init__(filepath_list, img_shape, color_mode, resize_method,
                         preserve_aspect_ratio, dtype, preprocessing_function)
        self.shape = (self.shape[0], np.prod(self.shape[1:]))

    def __getitem__(self, idx):
        return super().__getitem__(idx).flatten()



"""Custom data ingestion, including out-of-memory training by inheriting
from the IndexableDataset class."""

from .image import ImageDataset, FlattenedImageDataset
from .sequence import IndexableDataset



"""Supervised neighbour retrieval by label"""
from collections.abc import Sequence
import numpy as np


class LabeledNeighbourMap(Sequence):
    """Retrieves neighbour indices according to class labels provided in constructor.
    Rows with the same label will be regarded as neighbours."""
    def __init__(self, labels):
        """Constructs a LabeledNeighbourMap instance from a list of labels.
        :param labels list: List of labels for each data-point. One label per data-point."""
        class_indicies = {label: np.argwhere(labels == label).ravel()
                          for label in np.unique(labels)}
        neighbour_matrix = np.array([class_indicies[label] for label in labels])
        self.neighbour_matrix = neighbour_matrix
    def __len__(self):
        """Returns the number of rows in the data"""
        return len(self.neighbour_matrix)
    def __getitem__(self, idx):
        """Retrieves the neighbours for the row index provided"""
        return self.neighbour_matrix[idx]



""" Triplet loss functions for training a siamese network with three subnetworks.
    All loss function variants are accessible through the `triplet_loss` function
    by specifying the distance as a string.
"""

import functools
from tensorflow.keras import backend as K
import tensorflow as tf
from tensorflow import keras
import numpy as np

from .distances import euclidean_distance, manhattan_distance, chebyshev_distance, cosine_distance


loss_dict = {}
def register_loss(loss_fn=None, *, name=None):
    """Registers a class definition or Callable as an ivis loss function.
    A mapping will be created between the name and the loss function passed.
    If a class definition is provided, an instance will be created, passing the name
    as an argument.

    If no name is provided to this function, the name of the passed function will be used
    as a key.

    The loss function must have two parameters, (y_true, y_pred)
    and calculates the loss for a batch of triplet inputs (y_pred).
    y_pred is expected to be of shape: (3, batch_size, embedding_dims).

    Usage:
        .. code-block:: python

            @register_loss
            def custom_loss(y_true, y_pred):
                pass
            model = Ivis(distance='custom_loss')"""

    if loss_fn is None:
        return functools.partial(register_loss, name=name)

    key = name or loss_fn.__name__
    if isinstance(loss_fn, type):
        loss_dict[key] = loss_fn(name=key)
    else:
        loss_dict[key] = loss_fn
    return loss_fn


def triplet_loss(distance='pn'):
    """Returns a previously registered triplet loss function associated
    with the string 'distance'. If passed a callable, just returns it."""
    if callable(distance):
        return distance
    try:
        loss_fn = loss_dict[distance]
        return loss_fn
    except KeyError:
        raise ValueError("Loss function {} not registered with ivis".format(distance))


@register_loss(name='pn')
class EuclideanPnLoss:
    """Calculates the pn loss (a variant of triplet loss) between anchor, positive and negative
    examples in a triplet based on euclidean distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)

        anchor_positive_distance = euclidean_distance(anchor, positive)
        anchor_negative_distance = euclidean_distance(anchor, negative)
        positive_negative_distance = euclidean_distance(positive, negative)

        stacked_an_pn_distance = [anchor_negative_distance, positive_negative_distance]

        return K.mean(K.maximum(anchor_positive_distance - stacked_an_pn_distance + self.margin, 0))


@register_loss(name='euclidean')
class EuclideanTripletLoss:
    """Calculates the standard triplet loss between anchor, positive and negative
    examples in a triplet based on euclidean distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)
        return K.mean(K.maximum(euclidean_distance(anchor, positive) - euclidean_distance(anchor, negative) + self.margin, 0))


@register_loss(name='manhattan_pn')
class ManhattanPnLoss:
    """Calculates the pn loss (a variant of triplet loss) between anchor, positive and negative
    examples in a triplet based on manhattan distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)

        anchor_positive_distance = manhattan_distance(anchor, positive)
        anchor_negative_distance = manhattan_distance(anchor, negative)
        positive_negative_distance = manhattan_distance(positive, negative)

        stacked_an_pn_distance = [anchor_negative_distance, positive_negative_distance]

        return K.mean(K.maximum(anchor_positive_distance - stacked_an_pn_distance + self.margin, 0))


@register_loss(name='manhattan')
class ManhattanTripletLoss:
    """Calculates the standard triplet loss between anchor, positive and negative
    examples in a triplet based on manhattan distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)
        return K.mean(K.maximum(manhattan_distance(anchor, positive) - manhattan_distance(anchor, negative) + self.margin, 0))


@register_loss(name='chebyshev_pn')
class ChebyshevPnLoss:
    """Calculates the pn loss (a variant of triplet loss) between anchor, positive and negative
    examples in a triplet based on chebyshev distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)

        anchor_positive_distance = chebyshev_distance(anchor, positive)
        anchor_negative_distance = chebyshev_distance(anchor, negative)
        positive_negative_distance = chebyshev_distance(positive, negative)

        stacked_an_pn_distance = [anchor_negative_distance, positive_negative_distance]

        return K.mean(K.maximum(anchor_positive_distance - stacked_an_pn_distance + self.margin, 0))


@register_loss(name='chebyshev')
class ChebyshevTripletLoss:
    """Calculates the standard triplet loss between anchor, positive and negative
    examples in a triplet based on chebyshev distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)
        return K.mean(K.maximum(chebyshev_distance(anchor, positive) - chebyshev_distance(anchor, negative) + self.margin, 0))


@register_loss(name='cosine_pn')
class CosinePnLoss:
    """Calculates the pn loss (a variant of triplet loss) between anchor, positive and negative
    examples in a triplet based on cosine distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)

        anchor_positive_distance = cosine_distance(anchor, positive)
        anchor_negative_distance = cosine_distance(anchor, negative)
        positive_negative_distance = cosine_distance(positive, negative)

        stacked_an_pn_distance = [anchor_negative_distance, positive_negative_distance]

        return K.mean(K.maximum(anchor_positive_distance - stacked_an_pn_distance + self.margin, 0))


@register_loss(name='cosine')
class CosineTripletLoss:
    """Calculates the standard triplet loss between anchor, positive and negative
    examples in a triplet based on cosine distance."""
    def __init__(self, margin=1, name=None):
        self.margin = margin
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)
        return K.mean(K.maximum(cosine_distance(anchor, positive) - cosine_distance(anchor, negative) + self.margin, 0))


@register_loss(name='softmax_ratio_pn')
class EuclideanSoftmaxRatioPnLoss:
    """Calculates a pn variant of the softmax ratio between anchor, positive
    and negative examples in a triplet based on euclidean distance."""
    def __init__(self, name=None):
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)

        anchor_positive_distance = euclidean_distance(anchor, positive)
        anchor_negative_distance = euclidean_distance(anchor, negative)
        positive_negative_distance = euclidean_distance(positive, negative)

        ideal_distance = K.constant([0., 1.])
        minimum_distance = K.min([anchor_negative_distance, positive_negative_distance], axis=0)

        softmax = K.softmax(K.concatenate([anchor_positive_distance, minimum_distance]))
        return K.mean(K.abs(ideal_distance - softmax))


@register_loss(name='softmax_ratio')
class EuclideanSoftmaxRatioLoss:
    """Calculates the standard softmax ratio between anchor, positive and negative
    examples in a triplet based on euclidean distance."""
    def __init__(self, name=None):
        name = name or self.__class__.__name__
        self.__name__ = name
    def __call__(self, y_true, y_pred):
        anchor, positive, negative = tf.unstack(y_pred)
        anchor_positive_distance = euclidean_distance(anchor, positive)
        anchor_negative_distance = euclidean_distance(anchor, negative)

        softmax = K.softmax(K.concatenate([anchor_positive_distance, anchor_negative_distance]))
        ideal_distance = K.constant([0., 1.])
        return K.mean(K.abs(ideal_distance - softmax))


def semi_supervised_loss(loss_function):
    """Wraps the provided ivis supervised loss function to deal with the partially
    labeled data. Returns a new 'semi-supervised' loss function that masks the
    loss on examples where label information is missing.

    Missing labels are assumed to be marked with -1."""

    @functools.wraps(loss_function)
    def new_loss_function(y_true, y_pred):
        mask = tf.cast(~tf.math.equal(y_true, -1), tf.float32)
        y_true_pos = tf.nn.relu(y_true)
        loss = loss_function(y_true_pos, y_pred)
        masked_loss = loss * mask
        return masked_loss

    return new_loss_function

def is_hinge(supervised_loss):
    loss = keras.losses.get(supervised_loss)
    if loss in get_hinge_losses():
        return True
    return False


def get_hinge_losses():
    hinge_losses = set([keras.losses.hinge,
                        keras.losses.squared_hinge,
                        keras.losses.categorical_hinge])
    return hinge_losses


def get_categorical_losses():
    categorical_losses = set([keras.losses.sparse_categorical_crossentropy,
                              keras.losses.categorical_crossentropy,
                              keras.losses.categorical_hinge,
                              keras.losses.binary_crossentropy,
                              keras.losses.hinge,
                              keras.losses.squared_hinge])
    return categorical_losses


def get_multiclass_losses():
    multiclass_losses = set([keras.losses.sparse_categorical_crossentropy,
                             keras.losses.categorical_crossentropy,
                             keras.losses.categorical_hinge])
    return multiclass_losses


def is_categorical(supervised_loss):
    loss = keras.losses.get(supervised_loss)
    if loss in get_categorical_losses():
        return True
    return False


def is_multiclass(supervised_loss):
    loss = keras.losses.get(supervised_loss)
    if loss in get_multiclass_losses():
        return True
    return False


def validate_sparse_labels(Y):
    if not zero_indexed(Y):
        raise ValueError('Ensure that your labels are zero-indexed')
    if not _consecutive_indexed(Y):
        raise ValueError('Ensure that your labels are indexed consecutively')


def zero_indexed(Y):
    if min(abs(Y)) != 0:
        return False
    return True


def _consecutive_indexed(Y):
    """ Assumes that Y is zero-indexed. """
    n_classes = len(np.unique(Y[Y != np.array(-1)]))
    if max(Y) >= n_classes:
        return False
    return True



import tensorflow as tf
from tensorflow.keras import backend as K

def euclidean_distance(x, y):
    return K.sqrt(K.maximum(K.sum(K.square(x - y), axis=-1, keepdims=True), K.epsilon()))


def manhattan_distance(x, y):
    return K.sum(K.abs(x - y), axis=-1, keepdims=True)


def chebyshev_distance(x, y):
    return K.max(K.abs(x - y), axis=-1, keepdims=True)


def cosine_distance(x, y):
    return K.sum(tf.nn.relu(1 + tf.keras.losses.cosine_similarity(x, y)))
